"""
Things needed for checking for updates.
"""
import json

import certifi
import urllib3
from packaging.version import Version

from unidown import static_data


def get_newest_app_version() -> Version:
    """
    Download the version tag from remote.

    :return: version from remote
    """
    with urllib3.PoolManager(cert_reqs='CERT_REQUIRED', ca_certs=certifi.where()) as p_man:
        pypi_json = p_man.urlopen('GET', static_data.PYPI_JSON_URL).data.decode('utf-8')
    releases = json.loads(pypi_json).get('releases', [])
    online_version = Version('0.0.0')
    for release in releases:
        cur_version = Version(release)
        if not cur_version.is_prerelease:
            online_version = max(online_version, cur_version)
    return online_version


def check_for_app_updates() -> bool:
    """
    Check for updates.

    :return: is update available
    """
    return get_newest_app_version() > Version(static_data.VERSION)



from __future__ import annotations

from datetime import datetime

from packaging.version import InvalidVersion, Version

from unidown.plugin.link_item import LinkItem
from unidown.plugin.link_item_dict import LinkItemDict
from unidown.plugin.plugin_info import PluginInfo


class SaveState:
    """
    Savestate of a plugin.

    :param version: savestate version
    :param plugin_info: plugin info
    :param last_update: last udpate time of the referenced data
    :param link_items: data
    :param version: savestate version

    :cvar time_format: time format to use
    :ivar version: savestate version
    :ivar plugin_info: plugin info
    :ivar last_update: newest udpate time
    :ivar link_items: data
    """
    # current savestate version which will be used
    time_format: str = "%Y%m%dT%H%M%S.%fZ"

    def __init__(self, plugin_info: PluginInfo, last_update: datetime, link_items: LinkItemDict,
                 version: Version = Version('1')):
        self.version: Version = version
        self.plugin_info: PluginInfo = plugin_info
        self.last_update: datetime = last_update
        self.link_items: LinkItemDict = link_items

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, self.__class__):
            return False
        return (self.plugin_info == other.plugin_info and self.link_items == other.link_items and self.version == other.version and
                self.last_update == other.last_update)

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    @classmethod
    def from_json(cls, data: dict) -> SaveState:
        """
        :param data: json data as dict
        :return: the SaveState
        :raises ValueError: version of SaveState does not exist or is empty
        :raises ~packaging.version.InvalidVersion: version is not PEP440 conform
        """
        data_dict = LinkItemDict()
        if 'linkItems' not in data:
            raise ValueError("linkItems of SaveState does not exist.")
        for key, link_item in data['linkItems'].items():
            data_dict[key] = LinkItem.from_json(link_item)
        if 'meta' not in data or 'version' not in data['meta'] or data['meta']['version'] == "":
            raise ValueError("version of SaveState does not exist or is empty.")
        try:
            version = Version(data['meta']['version'])
        except InvalidVersion:
            raise InvalidVersion(f"Savestate version is not PEP440 conform: {data['meta']['version']}")
        return cls(PluginInfo.from_json(data['pluginInfo']), datetime.strptime(data['lastUpdate'], SaveState.time_format), data_dict, version)

    def to_json(self) -> dict:
        """
        Create json data.

        :return: json dictionary
        """
        result = {
            'meta': {'version': str(self.version)},
            'pluginInfo': self.plugin_info.to_json(),
            'lastUpdate': self.last_update.strftime(SaveState.time_format),
            'linkItems': {},
        }
        for key, link_item in self.link_items.items():
            result['linkItems'][key] = link_item.to_json()
        return result

    def upgrade(self) -> SaveState:
        """
        Upgrading current savestate to the latest savestate version.

        :return: upgraded savestate
        """
        return self



"""
Default exceptions of plugins.
"""


class PluginException(Exception):
    """
    Base class for exceptions in a plugin.
    If catching this, it implicit that the plugin is unable to work further.

    :param msg: message

    :ivar msg: exception message
    """

    def __init__(self, msg: str = ''):
        super().__init__(msg)
        self.msg: str = msg



"""
Static and important data which is needed by the program and do not need any third party libraries. (this is important
because it is used inside the setup.py **| do not edit**.
"""

#: name of this program
NAME = "unidown"
#: long name of this program
LONG_NAME = "Universal Downloader"
#: version in PEP440 format
VERSION = '2.0.0.dev7'
#: url to the unidown pypi json
PYPI_JSON_URL = 'https://pypi.python.org/pypi/unidown/json'
#: author
AUTHOR = 'Iceflower S'
#: author github link
AUTHOR_GITHUB = 'https://github.com/IceflowRE'
#: author email
AUTHOR_EMAIL = 'iceflower@gmx.de'
#: short description
DESCRIPTION = 'Universal downloader, a modular extensible downloader who manage progress and updates.'
#: project url
PROJECT_URL = 'https://github.com/IceflowRE/unidown'



from __future__ import annotations

from datetime import datetime


class LinkItem:
    """
    Item which represents the data, who need to be downloaded. Has a name and an update time.

    :param name: name
    :param time: update time
    :raises ValueError: name cannot be empty or None
    :raises ValueError: time cannot be empty or None

    :cvar time_format: time format to use
    :ivar _name: name of the item
    :ivar _time: time of the item
    """
    time_format: str = "%Y%m%dT%H%M%S.%fZ"

    def __init__(self, name: str, time: datetime):
        self._name: str = ""
        self._time: datetime = datetime(1970, 1, 1)
        self.name = name
        self.time = time

    @classmethod
    def from_json(cls, data: dict) -> LinkItem:
        """
        Constructor from json dict.

        :param data: json data as dict
        :return: the LinkItem
        :raises ValueError: missing parameter
        """
        if 'name' not in data:
            raise ValueError("name is missing")
        if 'time' not in data:
            raise ValueError("time is missing")
        return cls(data['name'], datetime.strptime(data['time'], LinkItem.time_format))

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, self.__class__):
            return False
        return self._name == other._name and self._time == other._time

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def __repr__(self) -> str:
        return self.__str__()

    def __str__(self) -> str:
        return '(' + str(self._name) + ', ' + str(self._time) + ')'

    @property
    def name(self) -> str:
        """
        Plain getter.
        """
        return self._name

    @name.setter
    def name(self, name: str):
        """
        :raises ValueError: name cannot be empty or None
        """
        if name is None or name == '':
            raise ValueError("name cannot be empty or None.")
        self._name = name

    @property
    def time(self) -> datetime:
        """
        Plain getter.
        """
        return self._time

    @time.setter
    def time(self, time: datetime):
        """
        :raises ValueError: time cannot be None
        """
        if time is None:
            raise ValueError("time cannot be None.")
        self._time = time

    def to_json(self) -> dict:
        """
        Create json data.

        :return: json dictionary
        """
        return {'name': self._name, 'time': self._time.strftime(LinkItem.time_format)}



"""
Entry into the program.
"""
import argparse
import sys
import traceback
from argparse import ArgumentParser
from pathlib import Path

from unidown import static_data, tools
from unidown.core import manager
from unidown.core.settings import Settings
from unidown.plugin.a_plugin import APlugin


class PluginListAction(argparse.Action):
    """
    Lists all plugins which are available. Extension for the argparser.
    """

    def __init__(self, option_strings, dest, **kwargs):
        super().__init__(option_strings, dest, nargs=0, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        tools.print_plugin_list(APlugin.get_plugins())
        parser.exit()


def main(argv=None):
    """
    Entry point into the program. Gets the arguments from the console and proceed them with :class:`~argparse.ArgumentParser`.
    Returns if its success successful 0 else 1.
    """
    if sys.version_info[0] < 3 or sys.version_info[1] < 8:
        sys.exit('Only Python 3.8 or greater is supported. You are using:' + sys.version)

    if argv is None:
        argv = sys.argv[1:]

    parser = ArgumentParser(prog=static_data.LONG_NAME, description=static_data.DESCRIPTION)
    parser.add_argument('-v', '--version', action='version', version=(static_data.NAME + ' ' + static_data.VERSION))
    parser.add_argument('--list-plugins', action=PluginListAction, help="show plugin list and exit")

    parser.add_argument('-p', '--plugin', dest='plugin', default="", type=str, required=True, metavar='name',
                        help='plugin to execute')
    parser.add_argument('-o', '--option', action='append', nargs='+', dest='options', type=str, metavar='option',
                        help='options passed to the plugin, e.g. `-o username=South American coati -o password=Nasua Nasua`')
    parser.add_argument('-r', '--root', dest='root_dir', default=None, type=str, metavar='path',
                        help='main directory where all files will be created (default: %(default)s)')
    parser.add_argument('--logfile', dest='logfile', default=None, type=str, metavar='path',
                        help='log filepath relativ to the main dir (default: %(default)s)')
    parser.add_argument('-l', '--log', dest='log_level', choices=['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'], default='INFO',
                        help='set the logging level (default: %(default)s)')

    args = parser.parse_args(argv)
    try:
        root_dir = args.root_dir
        if args.root_dir is not None:
            root_dir = Path(args.root_dir)
        log_file = args.logfile
        if args.logfile is not None:
            log_file = Path(args.logfile)
        settings = Settings(root_dir, log_file, args.log_level)
        manager.init_logging(settings)
    except PermissionError:
        print('Cant create needed folders. Make sure you have write permissions.')
        sys.exit(1)
    except FileExistsError as ex:
        print(ex)
        sys.exit(1)
    except Exception as ex:
        print('Something went wrong: ' + traceback.format_exc(ex.__traceback__))
        sys.exit(1)
    manager.check_update()
    manager.run(settings, args.plugin, args.options)
    manager.shutdown()
    sys.exit(0)



from unidown.plugin.a_plugin import APlugin
from unidown.plugin.exceptions import PluginException
from unidown.plugin.link_item import LinkItem
from unidown.plugin.link_item_dict import LinkItemDict
from unidown.plugin.plugin_info import PluginInfo
from unidown.plugin.savestate import SaveState

__all__ = ["APlugin", "PluginException", "LinkItem", "LinkItemDict", "PluginInfo", "SaveState"]



"""
Different tools.
"""

from pathlib import Path
from typing import Dict

import pkg_resources


def unlink_dir_rec(path: Path):
    """
    Delete a folder recursive.

    :param path: folder to deleted
    """
    if not path.exists() or not path.is_dir():
        return
    for sub in path.iterdir():
        if sub.is_dir():
            unlink_dir_rec(sub)
        else:
            sub.unlink()
    path.rmdir()


def print_plugin_list(plugins: Dict[str, pkg_resources.EntryPoint]):
    """
    Prints all registered plugins and checks if they can be loaded or not.

    :param plugins: plugins name to entrypoint
    """
    for trigger, entry_point in plugins.items():
        try:
            plugin_class = entry_point.load()
            version = str(plugin_class._info.version)
            print(
                f"{trigger} (ok)\n"
                f"    {version}"
            )
        except Exception:
            print(f"{trigger} (failed)")



import multiprocessing
from pathlib import Path


class Settings:
    """
    :ivar _root_dir: root path
    :ivar temp_dir: temporary main path, here are the sub folders for every plugin
    :ivar download_dir: download main path, here are the sub folders for every plugin
    :ivar savestate_dir: savestates main path, here are the sub folders for every plugin
    :ivar log_file: log file of the program
    :ivar available_plugins: available plugins which are found at starting the program, name -> EntryPoint
    :ivar using_cores: how many _cores should be used
    :ivar _log_level: log level
    :ivar _disable_tqdm: if the console progress bar is disabled

    :param root_dir: root dir
    :param log_file: log file
    """

    def __init__(self, root_dir: Path = None, log_file: Path = None, log_level: str = 'INFO'):
        if root_dir is None:
            root_dir = Path('./')
        if log_file is None:
            log_file = root_dir.joinpath(Path('unidown.log'))
        self._root_dir: Path = root_dir
        self._temp_dir: Path = self._root_dir.joinpath(Path('temp/'))
        self._download_dir: Path = self._root_dir.joinpath(Path('downloads/'))
        self._savestate_dir: Path = self._root_dir.joinpath(Path('savestates/'))
        self._log_file: Path = log_file
        self._cores = min(4, max(1, multiprocessing.cpu_count() - 1))
        self._log_level = log_level
        self._disable_tqdm = False

    def mkdir(self):
        """
        Create all base directories.
        """
        self._root_dir.mkdir(parents=True, exist_ok=True)
        self._temp_dir.mkdir(parents=True, exist_ok=True)
        self._download_dir.mkdir(parents=True, exist_ok=True)
        self._savestate_dir.mkdir(parents=True, exist_ok=True)

    def check_dirs(self):
        """
        Check the directories if they exist.

        :raises FileExistsError: if a file exists but is not a directory
        """
        dirs = [self._root_dir, self._temp_dir, self._download_dir, self._savestate_dir]
        for directory in dirs:
            if directory.exists() and not directory.is_dir():
                raise FileExistsError(str(directory.resolve()) + " cannot be used as a directory.")

    @property
    def root_dir(self) -> Path:
        """
        Plain getter.
        """
        return self._root_dir

    @property
    def temp_dir(self) -> Path:
        """
        Plain getter.
        """
        return self._temp_dir

    @property
    def download_dir(self) -> Path:
        """
        Plain getter.
        """
        return self._download_dir

    @property
    def savestate_dir(self) -> Path:
        """
        Plain getter.
        """
        return self._savestate_dir

    @property
    def log_file(self) -> Path:
        """
        Plain getter.
        """
        return self._log_file

    @property
    def cores(self) -> int:
        """
        Plain getter.
        """
        return self._cores

    @property
    def log_level(self) -> str:
        """
        Plain getter.
        """
        return self._log_level

    @property
    def disable_tqdm(self) -> bool:
        """
        Plain getter.
        """
        return self._disable_tqdm



import json
import logging
import time
from abc import ABC, abstractmethod
from concurrent.futures import ThreadPoolExecutor, as_completed
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, Tuple

import certifi
import pkg_resources
import urllib3
import urllib3.util
from packaging.version import Version
from tqdm import tqdm
from urllib3.exceptions import HTTPError

from unidown import tools
from unidown.core.settings import Settings
from unidown.plugin.exceptions import PluginException
from unidown.plugin.link_item_dict import LinkItemDict
from unidown.plugin.plugin_info import PluginInfo
from unidown.plugin.savestate import SaveState


class APlugin(ABC):
    """
    Abstract class of a plugin. Provides all needed variables and methods.

    :param options: parameters which can included optional parameters
    :raises ~unidown.plugin.exceptions.PluginException: can not create default plugin paths

    :cvar _info: information about the plugin
    :cvar _savestate_cls: savestate class to use
    :ivar _disable_tqdm: if the tqdm progressbar should be disabled **| do not edit**
    :ivar _log: use this for logging **| do not edit**
    :ivar _simul_downloads: number of simultaneous downloads
    :ivar _temp_path: path where the plugin can place all temporary data **| do not edit**
    :ivar _download_path: general download path of the plugin **| do not edit**
    :ivar _savestate_file: file which contains the latest savestate of the plugin **| do not edit**
    :ivar _last_update: latest update time of the referencing data **| do not edit**
    :ivar _unit: the thing which should be downloaded, may be displayed in the progress bar
    :ivar _download_data: referencing data **| do not edit**
    :ivar _downloader: downloader which will download the data **| do not edit**
    :ivar _savestate: savestate of the plugin
    :ivar _options: options which the plugin uses internal, should be used for the given options at init
    """
    _info: PluginInfo = None
    _savestate_cls = SaveState

    def __init__(self, settings: Settings, options: Dict[str, Any] = None):
        if options is None:
            options = {}
        if self._info is None:
            raise ValueError("info is not set.")

        self._disable_tqdm = settings.disable_tqdm
        self._log: logging.Logger = logging.getLogger(self._info.name)
        self._simul_downloads: int = settings.cores

        self._temp_path: Path = settings.temp_dir.joinpath(self.name)
        self._download_path: Path = settings.download_dir.joinpath(self.name)
        self._savestate_file: Path = settings.savestate_dir.joinpath(self.name + '_save.json')

        try:
            settings.mkdir()
            self._temp_path.mkdir(parents=True, exist_ok=True)
            self._download_path.mkdir(parents=True, exist_ok=True)
            settings.savestate_dir.mkdir(parents=True, exist_ok=True)
        except PermissionError:
            raise PluginException('Can not create default plugin paths, due to a permission error.')

        # cached data
        self._last_update: datetime = datetime(1970, 1, 1)
        self._download_data: LinkItemDict = LinkItemDict()

        self._savestate: SaveState = self._savestate_cls(self.info, self.last_update, LinkItemDict())

        self._unit: str = 'item'
        self._downloader: urllib3.HTTPSConnectionPool = urllib3.HTTPSConnectionPool(
            self.info.host, maxsize=self._simul_downloads, cert_reqs='CERT_REQUIRED', ca_certs=certifi.where()
        )

        # load options
        self._options: Dict[str, Any] = options
        self._load_default_options()

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, self.__class__):
            return False
        return self.info == other.info

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    @property
    def log(self) -> logging.Logger:
        """
        Plain getter.
        """
        return self._log

    @property
    def simul_downloads(self) -> int:
        """
        Plain getter.
        """
        return self._simul_downloads

    @property
    def info(self) -> PluginInfo:
        """
        Plain getter.
        """
        return self._info

    @property
    def host(self) -> str:
        """
        Plain getter.
        """
        return self._info.host

    @property
    def name(self) -> str:
        """
        Plain getter.
        """
        return self._info.name

    @property
    def version(self) -> Version:
        """
        Plain getter.
        """
        return self._info.version

    @property
    def temp_path(self) -> Path:
        """
        Plain getter.
        """
        return self._temp_path

    @property
    def download_path(self) -> Path:
        """
        Plain getter.
        """
        return self._download_path

    @property
    def savestate(self):
        """
        Plain getter.
        """
        return self._savestate

    @property
    def last_update(self) -> datetime:
        """
        Plain getter.
        """
        return self._last_update

    @property
    def download_data(self) -> LinkItemDict:
        """
        Plain getter.
        """
        return self._download_data

    @property
    def unit(self) -> str:
        """
        Plain getter.
        """
        return self._unit

    @property
    def options(self) -> Dict[str, Any]:
        """
        Plain getter.
        """
        return self._options

    def load_savestate(self):
        """
        Load the save of the plugin.

        :raises ~unidown.plugin.exceptions.PluginException: broken savestate json
        :raises ~unidown.plugin.exceptions.PluginException: different savestate versions
        :raises ~unidown.plugin.exceptions.PluginException: different plugin versions
        :raises ~unidown.plugin.exceptions.PluginException: different plugin names
        :raises ~unidown.plugin.exceptions.PluginException: could not parse the json
        """
        if not self._savestate_file.exists():
            self.log.info("No savestate file found.")
            return

        with self._savestate_file.open(encoding="utf8") as reader:
            try:
                savestate_json = json.loads(reader.read())
            except Exception:
                raise PluginException(f"Broken savestate json. Please fix or delete this file (you may lose data in this case): {self._savestate_file}")

        try:
            savestate = self._savestate_cls.from_json(savestate_json)
        except Exception as ex:
            raise PluginException(f"Could not load savestate from json {self._savestate_file}: {ex}")
        else:
            del savestate_json
        savestate = self._savestate_cls.upgrade(savestate)

        if savestate.plugin_info.name != self.info.name:
            raise PluginException(
                "Save state plugin ({name}) does not match the current ({cur_name}).".format(name=savestate.plugin_info.name, cur_name=self.name))
        self._savestate = savestate

    @abstractmethod
    def _create_last_update_time(self) -> datetime:
        """
        Get the newest update time from the referencing data.
        **Has to be implemented inside Plugins.**

        :raises NotImplementedError: abstract method
        """
        raise NotImplementedError

    def update_last_update(self):
        """
        Call this to update the latest update time. Calls :func:`~unidown.plugin.a_plugin.APlugin._create_last_update_time`.
        """
        self._last_update = self._create_last_update_time()

    @abstractmethod
    def _create_download_links(self) -> LinkItemDict:
        """
        Get the download links in a specific format.
        **Has to be implemented inside Plugins.**

        :raises NotImplementedError: abstract method
        """
        raise NotImplementedError

    def update_download_links(self):
        """
        Update the download links. Calls :func:`~unidown.plugin.a_plugin.APlugin._create_download_links`.
        """
        self._download_data = self._create_download_links()

    def download(self, link_items: LinkItemDict, folder: Path, desc: str, unit: str):
        """
        .. warning::

            The parameters may change in future versions. (e.g. change order and accept another host)

        Download the given LinkItem dict from the plugins host, to the given path. Proceeded with multiple connections
        :attr:`~unidown.plugin.a_plugin.APlugin._simul_downloads`. After
        :func:`~unidown.plugin.a_plugin.APlugin.check_download` is recommend.

        This function don't use an internal `link_item_dict`, `delay` or `folder` directly set in options or instance
        vars, because it can be used aside of the normal download routine inside the plugin itself for own things.
        As of this it still needs access to the logger, so a staticmethod is not possible.

        :param link_items: data which gets downloaded
        :param folder: target download folder
        :param desc: description of the progressbar
        :param unit: unit of the download, shown in the progressbar
        """
        # TODO: add other optional host?
        if not link_items:
            return

        job_list = []
        with ThreadPoolExecutor(max_workers=self._simul_downloads) as executor:
            for link, item in link_items.items():
                job = executor.submit(self.download_as_file, link, folder.joinpath(item.name), self._options['delay'])
                job_list.append(job)

            pbar = tqdm(as_completed(job_list), total=len(job_list), desc=desc, unit=unit, mininterval=1, ncols=100, disable=self._disable_tqdm)
            for _ in pbar:
                pass

        for job in job_list:
            try:
                job.result()
            except HTTPError as ex:
                self.log.exception(f"Failed to download: {str(ex)}")

    def download_as_file(self, url: str, target_file: Path, delay: float = 0) -> str:
        """
        Download the given url to the given target folder.

        :param url: link
        :param target_file: target file
        :param delay: after download wait this much seconds
        :return: url
        :raises ~urllib3.exceptions.HTTPError: if the connection has an error
        """
        if target_file.exists():
            new_name = target_file
            while new_name.exists():
                new_name = new_name.with_name(new_name.stem + '_r' + ''.join(new_name.suffixes))
            target_file.rename(new_name)
            self.log.critical(f"target file exists! renaming '{target_file}' to '{new_name}'")

        with self._downloader.request('GET', url, preload_content=False, retries=urllib3.util.retry.Retry(3)) as reader:
            if reader.status == 200:
                with target_file.open(mode='wb') as writer:
                    writer.write(reader.data)
            else:
                raise HTTPError(f"{url} | {reader.status}")

        if delay > 0:
            time.sleep(delay)

        return url

    def check_download(self, link_item_dict: LinkItemDict, folder: Path, log: bool = True) -> Tuple[LinkItemDict, LinkItemDict]:
        """
        Check if the download of the given dict was successful. No proving if the content of the file is correct too.

        :param link_item_dict: dict which to check
        :param folder: folder where the downloads are saved
        :param log: if the lost items should be logged
        :return: succeeded and failed
        """
        succeed = LinkItemDict({link: item for link, item in link_item_dict.items() if folder.joinpath(item.name).is_file()})
        failed = LinkItemDict({link: item for link, item in link_item_dict.items() if link not in succeed})

        if failed and log:
            for link, item in failed.items():
                self.log.error(f"Not downloaded: {self.info.host + link} - {item.name}")

        return succeed, failed

    def update_savestate(self, new_items: LinkItemDict):
        """
        Update savestate.

        :param new_items: new items
        """
        self._savestate.plugin_info = self.info
        self._savestate.last_update = self.last_update
        self._savestate.link_items.actualize(new_items)

    def save_savestate(self):
        """
        Save meta data about the downloaded things and the plugin to file.
        """
        with self._savestate_file.open(mode='w', encoding="utf8") as writer:
            writer.write(json.dumps(self._savestate.to_json()))

    def clean_up(self):
        """
        Default clean up for a module.
        Deletes :attr:`~unidown.plugin.a_plugin.APlugin._temp_path`.
        """
        self._downloader.close()
        tools.unlink_dir_rec(self._temp_path)

    def delete_data(self):
        """
        Delete everything which is related to the plugin. **Do not use if you do not know what you do!**
        """
        self.clean_up()
        tools.unlink_dir_rec(self._download_path)
        self._savestate_file.unlink(missing_ok=True)

    def _load_default_options(self):
        """
        Loads default options if they were not passed at creation.
        """
        if 'delay' in self._options:
            try:
                self._options['delay'] = float(self._options['delay'])
            except ValueError:
                self._options['delay'] = 0
                self.log.warning(f"Plugin option 'delay' was not a float. Using {self._options['delay']}s.")
        else:
            self._options['delay'] = 0
            self.log.warning(f"Plugin option 'delay' is missing. Using {self._options['delay']}s.")

    @staticmethod
    def get_plugins() -> Dict[str, pkg_resources.EntryPoint]:
        """
        Get all available plugins for unidown.

        :return: plugin name list
        """
        return {entry.name: entry for entry in pkg_resources.iter_entry_points('unidown.plugin')}



from enum import IntEnum


class PluginState(IntEnum):
    """
    State of a plugin, after it ended or was not found.
    """
    EndSuccess = 0  #: successfully end
    RunFail = 1  #: raised an ~unidown.plugin.exceptions.PluginException
    RunCrash = 2  #: raised an exception but ~unidown.plugin.exceptions.PluginException
    LoadCrash = 3  #: raised an exception while loading/ initializing
    NotFound = 4  #: plugin was not found



from __future__ import annotations

from packaging.version import InvalidVersion, Version


class PluginInfo:
    """
    .. warning:

        Parameters may change in the future.

    Information about the module. Those information will be saved into the save files as well.

    :param name: the name of the plugin
    :param version: version, PEP440 conform
    :param host: host url of the main data
    :raises ValueError: name is empty
    :raises ValueError: host is empty
    :raises ~packaging.version.InvalidVersion: version is not PEP440 conform

    :ivar name: name of the plugin
    :ivar host: host url of the main data
    :ivar version: plugin version
    """

    def __init__(self, name: str, version: str, host: str):
        if name is None or name == "":
            raise ValueError("Plugin name cannot be empty.")
        if host is None or host == "":
            raise ValueError("Plugin host cannot be empty.")
        self.name: str = name
        self.host: str = host

        try:
            self.version: Version = Version(version)
        except InvalidVersion:
            raise InvalidVersion(f"Plugin version is not PEP440 conform: {version}")

    @classmethod
    def from_json(cls, data: dict) -> PluginInfo:
        """
        Constructor from json dict.

        :param data: json data as dict
        :return: the plugin info
        :raises ValueError: name is missing
        :raises ValueError: version is missing
        :raises ValueError: host is missing
        """
        if 'name' not in data:
            raise ValueError("name is missing")
        if 'version' not in data:
            raise ValueError("version is missing")
        if 'host' not in data:
            raise ValueError("host is missing")
        return cls(data['name'], data['version'], data['host'])

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, self.__class__):
            return False
        return self.name == other.name and self.host == other.host and self.version == other.version

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def __str__(self) -> str:
        return self.name + " - " + str(self.version) + " : " + self.host

    def to_json(self) -> dict:
        """
        Create json data.

        :return: json dictionary
        """
        return {'name': self.name, 'version': str(self.version), 'host': self.host}



"""
Manager of the whole program, contains the most important functions as well as the download routine.
"""
import logging
import multiprocessing
import platform
from typing import List, Dict, Any

from unidown import static_data
from unidown.core import updater
from unidown.core.plugin_state import PluginState
from unidown.core.settings import Settings
from unidown.plugin.a_plugin import APlugin
from unidown.plugin.exceptions import PluginException
from unidown.plugin.link_item_dict import LinkItemDict


def init_logging(settings: Settings):
    """
    Initialize the _downloader.

    :param settings: settings
    """
    logging.basicConfig(filename=str(settings.log_file), filemode='a', level=settings.log_level,
                        format='%(asctime)s.%(msecs)03d | %(levelname)s - %(name)s | %(module)s.%(funcName)s: %('
                               'message)s', datefmt='%Y.%m.%d %H:%M:%S')
    logging.captureWarnings(True)

    info = f"{static_data.NAME} {static_data.VERSION}\n\n" \
           f"System: {platform.system()} - {platform.version()} - {platform.machine()} - {multiprocessing.cpu_count()} cores\n" \
           f"Python: {platform.python_version()} - {' - '.join(platform.python_build())}\n" \
           f"Arguments: main={settings.root_dir.resolve()} | logfile={settings.log_file} | loglevel={settings.log_level}\n" \
           f"Using cores: {settings.cores}\n\n"

    settings.log_file.parent.mkdir(parents=True, exist_ok=True)
    with settings.log_file.open(mode='w', encoding="utf8") as writer:
        writer.write(info)


def shutdown():
    """
    Close and exit important things.
    """
    logging.shutdown()


def download_from_plugin(plugin: APlugin):
    """
    Download routine.

    1. Get plugin from the given name
    2. Get the last overall update time
    3. Load the savestate
    4. Compare last update time with the one from the savestate
    5. Get the download links
    6. Compare received links and their times with the savestate
    7. Clean up names, to eliminate duplicated
    8. Download new and newer links
    9. Check downloaded data
    10. Update savestate
    11. Save new savestate to file

    :param plugin: plugin
    """
    # get last update date
    plugin.log.info('Get last update')
    plugin.update_last_update()
    # load old save state
    plugin.load_savestate()
    if plugin.last_update <= plugin.savestate.last_update:
        plugin.log.info('No update. Nothing to do.')
        return
    # get download links
    plugin.log.info('Get download links')
    plugin.update_download_links()
    # compare with save state
    new_items = LinkItemDict.get_new_items(plugin.savestate.link_items, plugin.download_data)
    plugin.log.info(f"Compared with save state: {str(len(plugin.download_data))}")
    if not new_items:
        plugin.log.info('No new data. Nothing to do.')
        return
    # clean up saving names
    plugin.log.info(f"Clean up names.")
    new_items.clean_up_names()
    # download new/updated data
    plugin.log.info(f"Download new {plugin.unit}s: {len(new_items)}")
    plugin.download(new_items, plugin.download_path, f"Download new {plugin.unit}s", plugin.unit)
    # check which downloads are succeeded
    succeeded, _ = plugin.check_download(new_items, plugin.download_path)
    plugin.log.info(f"Downloaded: {len(succeeded)}/{len(new_items)}")
    # update savestate link_item_dict with succeeded downloads dict
    plugin.log.info('Update savestate')
    plugin.update_savestate(succeeded)
    # write new savestate
    plugin.log.info('Write savestate')
    plugin.save_savestate()


def run(settings: Settings, plugin_name: str, raw_options: List[List[str]]) -> PluginState:
    """
    Run a plugin so use the download routine and clean up after.

    :param settings: settings to use
    :param plugin_name: name of plugin
    :param raw_options: parameters which will be send to the plugin initialization
    :return: ending state
    """
    if raw_options is None:
        options = {}
    else:
        options = get_options(raw_options)

    available_plugins = APlugin.get_plugins()
    if plugin_name not in available_plugins:
        msg = f"Plugin {plugin_name} was not found."
        logging.error(msg)
        return PluginState.NotFound

    try:
        plugin_class = available_plugins[plugin_name].load()
        plugin = plugin_class(settings, options)
    except Exception as ex:
        msg = f"Plugin {plugin_name} crashed while loading."
        logging.exception(msg, ex)
        print(f"{msg} Check log for more information.")
        return PluginState.LoadCrash
    else:
        logging.info(f"Loaded plugin: {plugin_name}")

    try:
        download_from_plugin(plugin)
        plugin.clean_up()
    except PluginException as ex:
        msg = f"Plugin {plugin.name} stopped working. Reason: {'unknown' if (ex.msg == '') else ex.msg}"
        logging.error(msg)
        print(msg)
        return PluginState.RunFail
    except Exception as ex:
        msg = f"Plugin {plugin.name} crashed."
        logging.exception(msg, ex)
        print(f"{msg} Check log for more information.")
        return PluginState.RunCrash
    else:
        logging.info(f"{plugin.name} ends without errors.")
        return PluginState.EndSuccess


def get_options(options: List[List[str]]) -> Dict[str, Any]:
    """
    Convert the option list to a dictionary where the key is the option and the value is the related option.
    Is called in the init.

    :param options: options given to the plugin.
    :return: dictionary which contains the option key and values
    """
    plugin_options = {}
    for sub_options in options:
        option = ' '.join(sub_options)
        option_key, _, option_value = option.partition('=')

        if option_key == '' or option_value == '':
            logging.warning(f"'{option}' is not valid and will be ignored.")
            continue
        plugin_options[option_key] = option_value
    return plugin_options


def check_update():
    """
    Check for app updates and print/log them.
    """
    logging.info('Check for app updates.')
    try:
        update = updater.check_for_app_updates()
    except Exception:
        logging.exception('Check for updates failed.')
        return
    if update:
        msg = f"Update available! ({static_data.PROJECT_URL})"
        print(msg)
        logging.info(msg)
    else:
        logging.info("No update available.")



#!/usr/bin/env python
"""
Setup.py for unidown.
"""

from pathlib import Path

from setuptools import find_packages, setup

from unidown import static_data

# get long description
with Path('README.rst').open(mode='r', encoding='UTF-8') as reader:
    LONG_DESCRIPTION = reader.read()

setup(
    name=static_data.NAME,
    version=static_data.VERSION,
    description=static_data.DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/x-rst',
    author=static_data.AUTHOR,
    author_email=static_data.AUTHOR_EMAIL,
    license='GPLv3',
    url=static_data.PROJECT_URL,
    classifiers=[
        'Programming Language :: Python :: 3.8',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Development Status :: 4 - Beta',
        'Operating System :: OS Independent',
        'Intended Audience :: Developers',
        'Intended Audience :: End Users/Desktop',
        'Natural Language :: English',
        'Environment :: Console',
        # 'Environment :: X11 Applications :: Qt',
    ],
    keywords='modular downloader',
    packages=find_packages(include=['unidown', 'unidown.*']),
    python_requires='>=3.8',
    install_requires=[
        'urllib3[secure]==1.25.8',
        'tqdm==4.44.0',
        'packaging==20.3',
    ],
    extras_require={
        'dev': [
            'flake8==3.7.9',
            'pylint==2.4.4',
            'pyroma==2.6',
            'pytest==5.4.1',
            'pytest-cov==2.8.1',
            'Sphinx==2.4.4',
            'sphinx-autodoc-typehints==1.10.3',
            'sphinx_rtd_theme==0.4.3',
            'twine==3.1.1',
            'setuptools==46.1.3',
            'wheel==0.34.2',
        ],
    },
    zip_safe=True,
    entry_points={
        'console_scripts': [
            'unidown = unidown.main:main',
        ],
        # 'gui_scripts': [
        #    '???',
        # ],
    },
)



from __future__ import annotations

import logging
from pathlib import Path

from tqdm import tqdm


class LinkItemDict(dict):
    """
    LinkItem dictionary, acts as a wrapper for special methods and functions.
    """

    def actualize(self, new_data: LinkItemDict, log: logging.Logger = None):
        """
        Actualize dictionary like an ~dict.update does. Additionally adds logging support.

        :param new_data: the data used for updating
        :param log: logger
        """
        if log is not None:
            for link, item in new_data.items():
                if link in self.keys():
                    log.info('Actualize item: ' + link + ' | ' + str(self[link]) + ' -> ' + str(item))
        self.update(new_data)

    def clean_up_names(self):
        """
        Rename duplicated names with an additional ``_r``.
        """
        for cur_link, cur_item in self.items():
            restart = True
            while restart:
                restart = False
                for link, item in self.items():
                    if cur_item.name == item.name and cur_link != link:
                        tmp = Path(cur_item.name)
                        cur_item.name = tmp.stem + '_d' + ''.join(tmp.suffixes)
                        restart = True
                        break

    @staticmethod
    def get_new_items(old_data: LinkItemDict, new_data: LinkItemDict, disable_tqdm: bool = False) -> LinkItemDict:
        """
        Get the new items which are not existing or are newer as in the old data set.

        :param old_data: old data
        :param new_data: new data
        :param disable_tqdm: disables tqdm progressbar
        :return: new and updated link items
        """
        if not old_data:
            return new_data
        if not new_data:
            return LinkItemDict()

        updated_data = LinkItemDict()
        for link, link_item in tqdm(new_data.items(), desc="Compare with save", unit="item", mininterval=1, ncols=100, disable=disable_tqdm):
            if (link not in old_data) or (link_item.time > old_data[link].time):
                updated_data[link] = link_item

        return updated_data
